/*
 * lxcall-a.9.c
 * Copyright (C) 2011-2014, 2016, 2017, 2019, 2020, 2022, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <e4.h>
#include <fd.h>
#include <lxcall-defs.h>
#include <lxcall-inter.h>
#include <lxcall-types.h>
#include <lxcast.h>

static int deck_call(const struct x1f4_function_type *,
		     const struct x1f4_function_type *, int);
static int miss_call(void *, int, void *, struct x1f4_dispatch_type **);
static int slip_call(void **, void **, const struct x1f4_function_type *,
		     const struct x1f4_function_type *,
		     struct x1f4_linetext_type *, unsigned, int *);

static int
deck_call(const struct x1f4_function_type *function_data,
	  const struct x1f4_function_type *function_text, int lock)
{
    int *args;
    unsigned class, count, flags;

    count = function_data->count;

    flags = function_text->flags;

    if (flags & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
	if (function_data->flags & X1f4_E4_POST_TYPE) {
	    flags |= X1f4_E4_POST_TYPE;
	}
    }

    args = (int *) function_data->args;

    class = function_text->count;
    if (lock < class) {
	int *line;
	const int *plot;

	line = args;
	plot = function_text->args;

	line += lock;
	plot += lock;

	if (flags & X1f4_E4_POST_TYPE) {
	    if (function_text->flags & X1f4_E4_POST_TYPE) {
		line[count] = plot[class];
	    } else {
		line[count] = 0;
	    }
	}
    }

    return 0;
}


static int
miss_call(void *text, int success, void *output,
	  struct x1f4_dispatch_type **dispatch)
{
    *dispatch = NULL;

    return success;
}


static int
slip_call(void **input, void **track,
	  const struct x1f4_function_type *function_data,
	  const struct x1f4_function_type *function_text,
	  struct x1f4_linetext_type *linetext_data, unsigned norm, int *lock)
{
    int status = 0;

    do {
	struct x1f4_dispatch_type *dispatch_data;
	void *output, *subtext;

	output = input[0];

	subtext = (void *) linetext_data->context;

	((void **) track[-1])[X1f4_E4_RENDITION] = &linetext_data->function;

	/*
	 * this is required because for SIDE/SLIP functions, the args array
	 * in the passed function (the RENDITION one) has been substituted
	 * with the function_data array.  an array that is being overwritten
	 * by our caller.
	 * the substitution is carried out by _post_call/_lock_call.
	 */
	if (function_text->flags & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	    for (; norm; norm--) {
		status = deck_call(function_data, function_text, *lock);
		if (status) {
		    break;
		} else {
		    lock++;
		}
	    }
	    if (status) {
		break;
	    }
	} else {
	}

	dispatch_data = input[2];

	dispatch_data->back = NULL;
	dispatch_data->call = miss_call;

	dispatch_data->transfer.context = subtext;
	dispatch_data->transfer.function_data = &linetext_data->function;
	dispatch_data->transfer.input = track;
	dispatch_data->transfer.output = output;
    } while (0);

    return status;
}


/*
 * coming in the wake of a _libx1f4i0_lxcall_post_call call, it expects
 * linetext_data->function set properly, bits include
 * X1f4_LXCALL_POST_LINE, etc.
 */
int
_libx1f4i0_lxcall_into_call(struct x1f4_linetext_type *linetext_data,
			    const struct x1f4_function_type *function_data,
			    struct center_type *center_data,
			    unsigned norm, int *lock,
			    struct fxdata_type *fxdata_data,
			    struct x1f4_dxcast_type *dxcast_data)
{
    const int *hold;
    int status;
    unsigned i;
    void ***input, **track;

    input = center_data->input;

    track = input[1];

    center_data->input = &track;

    hold = function_data->args + function_data->count;

    i = norm;
    while (1) {
	status = _libx1f4i0_lxcall_even_mark
	    (center_data, center_data->function_text, function_data,
	     *lock, hold + *lock, fxdata_data, dxcast_data);
	if (status) {
	    break;
	} else {
	    i--;
	    if (i) {
		lock++;
		fxdata_data++;
		dxcast_data++;
	    } else {
		break;
	    }
	}
    }

    center_data->input = input;

    if (status) {
    } else {
	lock -= norm - 1;

	status = slip_call
	    ((void *) input, track, function_data, center_data->function_text,
	     linetext_data, norm, lock);
    }

    return status;
}
