/*
 * lxcall-a.a.c
 * Copyright (C) 2011-2014, 2016, 2017, 2019, 2020, 2022, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <e4.h>
#include <fd.h>
#include <lxcall-defs.h>
#include <lxcall-inter.h>
#include <lxcall-types.h>
#include <lxcast.h>

static int miss_call(void *, int, void *, struct x1f4_dispatch_type **);

static int
miss_call(void *text, int success, void *output,
	  struct x1f4_dispatch_type **dispatch)
{
    *dispatch = NULL;

    return success;
}


/*
 * coming in the wake of a _libx1f4i0_lxcall_post_call call, it expects
 * linetext_data->function set properly, bits include
 * X1f4_LXCALL_POST_LINE, etc.
 */
int
_libx1f4i0_lxcall_jack_call(struct x1f4_linetext_type *linetext_data,
			    const struct x1f4_function_type *function_data,
			    struct center_type *center_data,
			    int lock, const int *hold,
			    struct fxdata_type *fxdata_data,
			    struct x1f4_dxcast_type *dxcast_data)
{
    int status;
    void ***input, **track;

    input = center_data->input;

    track = input[1];

    center_data->input = &track;

    do {
	struct x1f4_dispatch_type *dispatch_data;
	void *output, *subtext;

	status = _libx1f4i0_lxcall_even_mark
	    (center_data, center_data->function_text, function_data,
	     lock, hold, fxdata_data, dxcast_data);
	if (status) {
	    break;
	}

	center_data->input = input;

	output = input[0];

	subtext = (void *) linetext_data->context;

	((void **) track[-1])[X1f4_E4_RENDITION] = &linetext_data->function;

	dispatch_data = (void *) input[2];

	dispatch_data->back = NULL;
	dispatch_data->call = miss_call;

	dispatch_data->transfer.context = subtext;
	dispatch_data->transfer.function_data = &linetext_data->function;
	dispatch_data->transfer.input = track;
	dispatch_data->transfer.output = output;
    } while (0);

    return status;
}
