/*
 * lxcall-a.b.c
 * Copyright (C) 2011-2014, 2016, 2017, 2019, 2020, 2022, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <e4.h>
#include <fd.h>
#include <lxcall-defs.h>
#include <lxcall-inter.h>
#include <lxcall-types.h>
#include <lxcast.h>

static int deck_call(const struct x1f4_function_type *,
		     const struct x1f4_function_type *, int);

static int
deck_call(const struct x1f4_function_type *function_data,
	  const struct x1f4_function_type *function_text, int lock)
{
    int *args;
    unsigned class, count, flags;

    count = function_data->count;

    flags = function_text->flags;

    if (flags & (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE)) {
	if (function_data->flags & X1f4_E4_POST_TYPE) {
	    flags |= X1f4_E4_POST_TYPE;
	}
    }

    args = (int *) function_data->args;

    class = function_text->count;
    if (lock < class) {
	int *line;
	const int *plot;

	line = args;
	plot = function_text->args;

	line += lock;
	plot += lock;

	if (flags & X1f4_E4_POST_TYPE) {
	    if (function_text->flags & X1f4_E4_POST_TYPE) {
		line[count] = plot[class];
	    } else {
		line[count] = 0;
	    }
	}
    }

    return 0;
}


/*
 * coming in the wake of a _libx1f4i0_lxcall_post_call call, it expects
 * linetext_data->function set properly, bits include
 * X1f4_LXCALL_POST_LINE, etc.
 */
int
_libx1f4i0_lxcall_rail_call(struct x1f4_linetext_type *linetext_data,
			    const struct x1f4_function_type *function_data,
			    void *output, struct center_type *center_data,
			    unsigned norm, int *lock,
			    struct fxdata_type *fxdata_data,
			    struct x1f4_dxcast_type *dxcast_data)
{
    const int *hold;
    int status;
    unsigned i;
    void **track;

    track = *center_data->input;

    hold = function_data->args + function_data->count;

    i = norm;
    while (1) {
	status = _libx1f4i0_lxcall_even_mark
	    (center_data, center_data->function_text, function_data,
	     *lock, hold + *lock, fxdata_data, dxcast_data);
	if (status) {
	    break;
	} else {
	    i--;
	    if (i) {
		lock++;
		fxdata_data++;
		dxcast_data++;
	    } else {
		break;
	    }
	}
    }
    if (status) {
    } else {
	do {
	    lock -= norm - 1;

	    ((void **) track[-1])[X1f4_E4_RENDITION] =
		&linetext_data->function;

	    /*
	     * this is required because for SIDE/SLIP functions, the args array
	     * in the passed function (the RENDITION one) has been substituted
	     * with the function_data array.  an array that is being
	     * overwritten by our caller.
	     * the substitution is carried out by _post_call/_lock_call.
	     */
	    if (center_data->function_text->flags
		& (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
		for (; norm; norm--) {
		    status = deck_call
			(function_data, center_data->function_text, *lock);
		    if (status) {
			break;
		    } else {
			lock++;
		    }
		}
		if (status) {
		    break;
		}
	    } else {
	    }

	    status = linetext_data->function.function
		((void *) linetext_data->context, output, track);
	} while (0);
    }

    return status;
}
